@@ -106,6 +106,13 @@ typedef struct {
         gpointer user_data;
 } CallbackData;
 
+typedef struct {
+        char *sid;
+        int seq;
+
+        xmlDoc *doc;
+} EmitNotifyData;
+
 static void
 subscribe_got_response (SoupSession       *session,
                         SoupMessage       *msg,
@@ -135,6 +142,32 @@ notify_data_free (NotifyData *data)
         g_slice_free (NotifyData, data);
 }
 
+/* Steals doc reference */
+static EmitNotifyData *
+emit_notify_data_new (const char *sid,
+                      int         seq,
+                      xmlDoc     *doc)
+{
+        EmitNotifyData *data;
+
+        data = g_slice_new (EmitNotifyData);
+
+        data->sid = g_strdup (sid);
+        data->seq = seq;
+        data->doc = doc;
+
+        return data;
+}
+
+static void
+emit_notify_data_free (EmitNotifyData *data)
+{
+        g_free (data->sid);
+        xmlFreeDoc (data->doc);
+
+        g_slice_free (EmitNotifyData, data);
+}
+
 static void
 gupnp_service_proxy_init (GUPnPServiceProxy *proxy)
 {
@@ -252,7 +285,7 @@ gupnp_service_proxy_dispose (GObject *object)
         }
         
         while (proxy->priv->pending_notifies) {
-                xmlFreeDoc (proxy->priv->pending_notifies->data);
+                emit_notify_data_free (proxy->priv->pending_notifies->data);
                 proxy->priv->pending_notifies =
                         g_list_delete_link (proxy->priv->pending_notifies,
                                             proxy->priv->pending_notifies);
@@ -295,7 +328,7 @@ gupnp_service_proxy_class_init (GUPnPServiceProxyClass *klass)
         g_type_class_add_private (klass, sizeof (GUPnPServiceProxyPrivate));
 
         /**
-         * GUPnPServiceProxy:subscribed
+         * GUPnPServiceProxy:subscribed:
          *
          * Whether we are subscribed to this service.
          **/
@@ -313,7 +346,7 @@ gupnp_service_proxy_class_init (GUPnPServiceProxyClass *klass)
                                        G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPServiceProxy::subscription-lost
+         * GUPnPServiceProxy::subscription-lost:
          * @proxy: The #GUPnPServiceProxy that received the signal
          * @error: A pointer to a #GError describing why the subscription has
          * been lost
@@ -336,7 +369,7 @@ gupnp_service_proxy_class_init (GUPnPServiceProxyClass *klass)
 }
 
 /**
- * gupnp_service_proxy_send_action
+ * gupnp_service_proxy_send_action:
  * @proxy: A #GUPnPServiceProxy
  * @action: An action
  * @error: The location where to store any error, or %NULL
@@ -378,7 +411,7 @@ stop_main_loop (GUPnPServiceProxy       *proxy,
 }
 
 /**
- * gupnp_service_proxy_send_action_valist
+ * gupnp_service_proxy_send_action_valist:
  * @proxy: A #GUPnPServiceProxy
  * @action: An action
  * @error: The location where to store any error, or %NULL
@@ -431,13 +464,14 @@ gupnp_service_proxy_send_action_valist (GUPnPServiceProxy *proxy,
 }
 
 /**
- * gupnp_service_proxy_send_action_hash
+ * gupnp_service_proxy_send_action_hash:
  * @proxy: A #GUPnPServiceProxy
  * @action: An action
  * @error: The location where to store any error, or %NULL
- * @in_hash: A #GHashTable of in parameter name and #GValue pairs
- * @out_hash: A #GHashTable of out parameter name and initialized
- * #GValue pairs
+ * @in_hash: (element-type utf8 GValue) (transfer none): A #GHashTable of in
+ * parameter name and #GValue pairs
+ * @out_hash: (inout) (element-type utf8 GValue) (transfer full): A #GHashTable
+ * of out parameter name and initialized #GValue pairs
  *
  * See gupnp_service_proxy_send_action(); this version takes a pair of
  * #GHashTable<!-- -->s for runtime determined parameter lists.
@@ -490,10 +524,82 @@ gupnp_service_proxy_send_action_hash (GUPnPServiceProxy *proxy,
 }
 
 /**
- * gupnp_service_proxy_begin_action
+ * gupnp_service_proxy_send_action_list:
+ * @proxy: (transfer none) : A #GUPnPServiceProxy
+ * @action: An action
+ * @error: The location where to store any error, or %NULL
+ * @in_names: (element-type utf8) (transfer none): #GList of 'in' parameter
+ * names (as strings)
+ * @in_values: (element-type GValue) (transfer none): #GList of values (as
+ * #GValue) that line up with @in_names
+ * @out_names: (element-type utf8) (transfer none): #GList of 'out' parameter
+ * names (as strings)
+ * @out_types: (element-type GType) (transfer none): #GList of types (as #GType)
+ * that line up with @out_names
+ * @out_values: (element-type GValue) (transfer full) (out): #GList of values
+ * (as #GValue) that line up with @out_names and @out_types.
+ *
+ * The synchronous variant of #gupnp_service_proxy_begin_action_list and
+ * #gupnp_service_proxy_end_action_list.
+ *
+ * Return value: %TRUE if sending the action was succesful.
+ **/
+gboolean
+gupnp_service_proxy_send_action_list (GUPnPServiceProxy *proxy,
+                                      const char        *action,
+                                      GError           **error,
+                                      GList             *in_names,
+                                      GList             *in_values,
+                                      GList             *out_names,
+                                      GList             *out_types,
+                                      GList            **out_values)
+{
+        GUPnPContext *context;
+        GMainContext *main_context;
+        GMainLoop *main_loop;
+        GUPnPServiceProxyAction *handle;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
+        g_return_val_if_fail (action, FALSE);
+
+        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (proxy));
+        main_context = gssdp_client_get_main_context (GSSDP_CLIENT (context));
+        main_loop = g_main_loop_new (main_context, TRUE);
+
+        handle = gupnp_service_proxy_begin_action_list (proxy,
+                                                        action,
+                                                        in_names,
+                                                        in_values,
+                                                        stop_main_loop,
+                                                        main_loop);
+        if (!handle) {
+                g_main_loop_unref (main_loop);
+
+                return FALSE;
+        }
+
+        /* Loop till we get a reply (or time out) */
+        if (g_main_loop_is_running (main_loop))
+                g_main_loop_run (main_loop);
+
+        g_main_loop_unref (main_loop);
+
+        if (!gupnp_service_proxy_end_action_list (proxy,
+                                                  handle,
+                                                  error,
+                                                  out_names,
+                                                  out_types,
+                                                  out_values))
+                return FALSE;
+
+        return TRUE;
+}
+
+/**
+ * gupnp_service_proxy_begin_action:
  * @proxy: A #GUPnPServiceProxy
  * @action: An action
- * @callback: The callback to call when sending the action has succeeded
+ * @callback: (scope async): The callback to call when sending the action has succeeded
  * or failed
  * @user_data: User data for @callback
  * @Varargs: tuples of in parameter name, in parameter type, and in parameter
@@ -504,8 +610,8 @@ gupnp_service_proxy_send_action_hash (GUPnPServiceProxy *proxy,
  * gupnp_service_proxy_end_action() to check for errors, to retrieve return
  * values, and to free the #GUPnPServiceProxyAction.
  *
- * Return value: A #GUPnPServiceProxyAction handle. This will
- * be freed when calling gupnp_service_proxy_cancel_action() or
+ * Return value: A #GUPnPServiceProxyAction handle. This will be freed when
+ * gupnp_service_proxy_cancel_action() or
  * gupnp_service_proxy_end_action_valist().
  **/
 GUPnPServiceProxyAction *
@@ -587,12 +693,11 @@ begin_action_msg (GUPnPServiceProxy              *proxy,
         /* Specify action */
         full_action = g_strdup_printf ("\"%s#%s\"", service_type, action);
         soup_message_headers_append (ret->msg->request_headers,
-				     "SOAPAction",
+                                     "SOAPAction",
                                      full_action);
         g_free (full_action);
 
-        /* Specify user agent and language */
-        http_request_set_user_agent (ret->msg);
+        /* Specify language */
         http_request_set_accept_language (ret->msg);
 
         /* Set up envelope */
@@ -638,8 +743,9 @@ action_got_response (SoupSession             *session,
                          "\"http://schemas.xmlsoap.org/soap/envelope/\"; ns=s");
 
                 /* Rename "SOAPAction" to "s-SOAPAction" */
-                full_action = soup_message_headers_get (msg->request_headers,
-                                                        "SOAPAction");
+                full_action = soup_message_headers_get_one
+                        (msg->request_headers,
+                         "SOAPAction");
                 soup_message_headers_append (msg->request_headers,
                                              "s-SOAPAction",
                                              full_action);
@@ -712,10 +818,10 @@ write_in_parameter (const char *arg_name,
 }
 
 /**
- * gupnp_service_proxy_begin_action_valist
+ * gupnp_service_proxy_begin_action_valist:
  * @proxy: A #GUPnPServiceProxy
  * @action: An action
- * @callback: The callback to call when sending the action has succeeded
+ * @callback: (scope async) : The callback to call when sending the action has succeeded
  * or failed
  * @user_data: User data for @callback
  * @var_args: A va_list of tuples of in parameter name, in parameter type, and
@@ -762,7 +868,8 @@ gupnp_service_proxy_begin_action_valist
                 arg_type = va_arg (var_args, GType);
                 g_value_init (&value, arg_type);
 
-                G_VALUE_COLLECT (&value, var_args, 0, &collect_error);
+                G_VALUE_COLLECT (&value, var_args, G_VALUE_NOCOPY_CONTENTS,
+                                 &collect_error);
                 if (!collect_error) {
                         write_in_parameter (arg_name, &value, ret->msg_str);
 
@@ -775,8 +882,8 @@ gupnp_service_proxy_begin_action_valist
                         g_free (collect_error);
 
                         /* we purposely leak the value here, it might not be
-	                 * in a sane state if an error condition occoured
-	                 */
+                         * in a sane state if an error condition occoured
+                         */
                 }
 
                 arg_name = va_arg (var_args, const char *);
@@ -792,10 +899,77 @@ gupnp_service_proxy_begin_action_valist
 }
 
 /**
- * gupnp_service_proxy_begin_action_hash
+ * gupnp_service_proxy_begin_action_list:
+ * @proxy: (transfer none): A #GUPnPServiceProxy
+ * @action: An action
+ * @in_names: (element-type utf8) (transfer none): #GList of 'in' parameter
+ * names (as strings)
+ * @in_values: (element-type GValue) (transfer none): #GList of values (as
+ * #GValue) that line up with @in_names
+ * @callback: (scope async) : The callback to call when sending the action has succeeded or
+ * failed
+ * @user_data: User data for @callback
+ *
+ * A variant of #gupnp_service_proxy_begin_action that takes lists of
+ * in-parameter names, types and values.
+ *
+ * Return value: (transfer none): A #GUPnPServiceProxyAction handle. This will
+ * be freed when calling #gupnp_service_proxy_cancel_action or
+ * #gupnp_service_proxy_end_action_list.
+ **/
+GUPnPServiceProxyAction *
+gupnp_service_proxy_begin_action_list
+                                   (GUPnPServiceProxy               *proxy,
+                                    const char                      *action,
+                                    GList                           *in_names,
+                                    GList                           *in_values,
+                                    GUPnPServiceProxyActionCallback  callback,
+                                    gpointer                         user_data)
+{
+        GUPnPServiceProxyAction *ret;
+        GList *names;
+        GList *values;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), NULL);
+        g_return_val_if_fail (action, NULL);
+        g_return_val_if_fail (callback, NULL);
+        g_return_val_if_fail (g_list_length (in_names) ==
+                              g_list_length (in_values),
+                              NULL);
+
+        /* Create message */
+        ret = begin_action_msg (proxy, action, callback, user_data);
+
+        if (ret->error) {
+                callback (proxy, ret, user_data);
+
+                return ret;
+        }
+
+        /* Arguments */
+        values = in_values;
+
+        for (names = in_names; names; names=names->next) {
+                GValue* val = values->data;
+
+                write_in_parameter (names->data,
+                                    val,
+                                    ret->msg_str);
+
+                values = values->next;
+        }
+
+        /* Finish and send off */
+        finish_action_msg (ret, action);
+
+        return ret;
+}
+
+/**
+ * gupnp_service_proxy_begin_action_hash:
  * @proxy: A #GUPnPServiceProxy
  * @action: An action
- * @callback: The callback to call when sending the action has succeeded
+ * @callback: (scope async): The callback to call when sending the action has succeeded
  * or failed
  * @user_data: User data for @callback
  * @hash: A #GHashTable of in parameter name and #GValue pairs
@@ -803,7 +977,7 @@ gupnp_service_proxy_begin_action_valist
  * See gupnp_service_proxy_begin_action(); this version takes a #GHashTable
  * for runtime generated parameter lists.
  *
- * Return value: A #GUPnPServiceProxyAction handle. This will
+ * Return value: (transfer none): A #GUPnPServiceProxyAction handle. This will
  * be freed when calling gupnp_service_proxy_cancel_action() or
  * gupnp_service_proxy_end_action_hash().
  **/
@@ -840,7 +1014,7 @@ gupnp_service_proxy_begin_action_hash
 }
 
 /**
- * gupnp_service_proxy_end_action
+ * gupnp_service_proxy_end_action:
  * @proxy: A #GUPnPServiceProxy
  * @action: A #GUPnPServiceProxyAction handle
  * @error: The location where to store any error, or %NULL
@@ -899,7 +1073,7 @@ check_action_response (GUPnPServiceProxy       *proxy,
         response = xmlRecoverMemory (action->msg->response_body->data,
                                      action->msg->response_body->length);
 
-	if (!response) {
+        if (!response) {
                 if (action->msg->status_code == SOUP_STATUS_OK) {
                         g_set_error (error,
                                      GUPNP_SERVER_ERROR,
@@ -914,7 +1088,7 @@ check_action_response (GUPnPServiceProxy       *proxy,
                 }
 
                 return NULL;
-	}
+        }
 
         /* Get parameter list */
         *params = xml_util_get_element ((xmlNode *) response,
@@ -1026,7 +1200,7 @@ read_out_parameter (const char *arg_name,
 }
 
 /**
- * gupnp_service_proxy_end_action_valist
+ * gupnp_service_proxy_end_action_valist:
  * @proxy: A #GUPnPServiceProxy
  * @action: A #GUPnPServiceProxyAction handle
  * @error: The location where to store any error, or %NULL
@@ -1051,6 +1225,7 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
 
         g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
         g_return_val_if_fail (action, FALSE);
+        g_return_val_if_fail (proxy == action->proxy, FALSE);
 
         /* Check for saved error from begin_action() */
         if (action->error) {
@@ -1107,14 +1282,99 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
 }
 
 /**
- * gupnp_service_proxy_end_action_hash
+ * gupnp_service_proxy_end_action_list:
  * @proxy: A #GUPnPServiceProxy
  * @action: A #GUPnPServiceProxyAction handle
  * @error: The location where to store any error, or %NULL
- * @hash: A #GHashTable of out parameter name and initialised #GValue pairs
+ * @out_names: (element-type utf8) (transfer none): #GList of 'out' parameter
+ * names (as strings)
+ * @out_types: (element-type GType) (transfer none): #GList of types (as #GType)
+ * that line up with @out_names
+ * @out_values: (element-type GValue) (transfer full) (out): #GList of values
+ * (as #GValue) that line up with @out_names and @out_types.
  *
- * See gupnp_service_proxy_end_action(); this version takes a #GHashTable
- * for runtime generated parameter lists.
+ * A variant of #gupnp_service_proxy_end_action that takes lists of
+ * out-parameter names, types and place-holders for values. The returned list
+ * in @out_values must be freed using #g_list_free and each element in it using
+ * #g_value_unset and #g_slice_free.
+ *
+ * Return value : %TRUE on success.
+ **/
+gboolean
+gupnp_service_proxy_end_action_list (GUPnPServiceProxy       *proxy,
+                                     GUPnPServiceProxyAction *action,
+                                     GError                 **error,
+                                     GList                   *out_names,
+                                     GList                   *out_types,
+                                     GList                  **out_values)
+{
+        xmlDoc *response;
+        xmlNode *params;
+        GList *names;
+        GList *types;
+        GList *out_values_list;
+
+        out_values_list = NULL;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
+        g_return_val_if_fail (action, FALSE);
+        g_return_val_if_fail (proxy == action->proxy, FALSE);
+
+        /* Check for saved error from begin_action() */
+        if (action->error) {
+                if (error)
+                        *error = action->error;
+                else
+                        g_error_free (action->error);
+
+                gupnp_service_proxy_action_free (action);
+
+                return FALSE;
+        }
+
+        /* Check response for errors and do initial parsing */
+        response = check_action_response (proxy, action, &params, error);
+        if (response == NULL) {
+                gupnp_service_proxy_action_free (action);
+
+                return FALSE;
+        }
+
+        /* Read arguments */
+        types = out_types;
+        for (names = out_names; names; names=names->next) {
+                GValue *val;
+
+                val = g_slice_new0 (GValue);
+                g_value_init (val, (GType) types->data);
+
+                read_out_parameter (names->data, val, params);
+
+                out_values_list = g_list_append (out_values_list, val);
+
+                types = types->next;
+        }
+
+        *out_values = out_values_list;
+
+        /* Cleanup */
+        gupnp_service_proxy_action_free (action);
+
+        xmlFreeDoc (response);
+
+        return TRUE;
+}
+
+/**
+ * gupnp_service_proxy_end_action_hash:
+ * @proxy: A #GUPnPServiceProxy
+ * @action: A #GUPnPServiceProxyAction handle
+ * @error: The location where to store any error, or %NULL
+ * @hash: (element-type utf8 GValue) (inout) (transfer none): A #GHashTable of
+ * out parameter name and initialised #GValue pairs
+ *
+ * See gupnp_service_proxy_end_action(); this version takes a #GHashTable for
+ * runtime generated parameter lists.
  *
  * Return value: %TRUE on success.
  **/
@@ -1129,6 +1389,7 @@ gupnp_service_proxy_end_action_hash (GUPnPServiceProxy       *proxy,
 
         g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
         g_return_val_if_fail (action, FALSE);
+        g_return_val_if_fail (proxy == action->proxy, FALSE);
 
         /* Check for saved error from begin_action() */
         if (action->error) {
@@ -1162,7 +1423,7 @@ gupnp_service_proxy_end_action_hash (GUPnPServiceProxy       *proxy,
 }
 
 /**
- * gupnp_service_proxy_cancel_action
+ * gupnp_service_proxy_cancel_action:
  * @proxy: A #GUPnPServiceProxy
  * @action: A #GUPnPServiceProxyAction handle
  *
@@ -1177,6 +1438,7 @@ gupnp_service_proxy_cancel_action (GUPnPServiceProxy       *proxy,
 
         g_return_if_fail (GUPNP_IS_SERVICE_PROXY (proxy));
         g_return_if_fail (action);
+        g_return_if_fail (proxy == action->proxy);
 
         if (action->msg != NULL) {
                 context = gupnp_service_info_get_context
@@ -1195,11 +1457,11 @@ gupnp_service_proxy_cancel_action (GUPnPServiceProxy       *proxy,
 }
 
 /**
- * gupnp_service_proxy_add_notify
+ * gupnp_service_proxy_add_notify:
  * @proxy: A #GUPnPServiceProxy
  * @variable: The variable to add notification for
  * @type: The type of the variable
- * @callback: The callback to call when @variable changes
+ * @callback: (scope async): The callback to call when @variable changes
  * @user_data: User data for @callback
  *
  * Sets up @callback to be called whenever a change notification for
@@ -1259,7 +1521,7 @@ gupnp_service_proxy_add_notify (GUPnPServiceProxy              *proxy,
 }
 
 /**
- * gupnp_service_proxy_remove_notify
+ * gupnp_service_proxy_remove_notify:
  * @proxy: A #GUPnPServiceProxy
  * @variable: The variable to add notification for
  * @callback: The callback to call when @variable changes
@@ -1325,80 +1587,130 @@ gupnp_service_proxy_remove_notify (GUPnPServiceProxy              *proxy,
         return found;
 }
 
+static void
+emit_notification (GUPnPServiceProxy *proxy,
+                   xmlNode           *var_node)
+{
+        NotifyData *data;
+        GValue value = {0, };
+        GList *l;
+
+        data = g_hash_table_lookup (proxy->priv->notify_hash, var_node->name);
+        if (data == NULL)
+                return;
+
+        /* Make a GValue of the desired type */
+        g_value_init (&value, data->type);
+
+        if (!gvalue_util_set_value_from_xml_node (&value, var_node)) {
+                g_value_unset (&value);
+
+                return;
+        }
+
+        /* Call callbacks */
+        for (l = data->callbacks; l; l = l->next) {
+                CallbackData *callback_data;
+
+                callback_data = l->data;
+
+                callback_data->callback (proxy,
+                                         (const char *) var_node->name,
+                                         &value,
+                                         callback_data->user_data);
+        }
+
+        /* Cleanup */
+        g_value_unset (&value);
+}
+
+static void
+emit_notifications_for_doc (GUPnPServiceProxy *proxy,
+                            xmlDoc            *doc)
+{
+        xmlNode *node;
+
+        node = xmlDocGetRootElement (doc);
+
+        /* Iterate over all provided properties */
+        for (node = node->children; node; node = node->next) {
+                xmlNode *var_node;
+
+                /* Although according to the UPnP specs, there should be only
+                 * one variable node inside a 'property' node, we still need to
+                 * entertain the possibility of multiple variables inside it to
+                 * be compatible with implementations using older GUPnP.
+                 */
+                for (var_node = node->children;
+                     var_node;
+                     var_node = var_node->next) {
+                        if (strcmp ((char *) node->name, "property") == 0)
+                                emit_notification (proxy, var_node);
+                }
+        }
+}
+
 /* Emit pending notifications. See comment below on why we do this. */
 static gboolean
 emit_notifications (gpointer user_data)
 {
         GUPnPServiceProxy *proxy = user_data;
+        GList *pending_notify;
+        gboolean resubscribe = FALSE;
 
         g_assert (user_data);
-        
-        while (proxy->priv->pending_notifies != NULL) {
-                xmlDoc *doc;
-                xmlNode *node;
 
-                doc = proxy->priv->pending_notifies->data;
-                node = xmlDocGetRootElement (doc);
-                
-                /* Iterate over all provided properties */
-                for (node = node->children; node; node = node->next) {
-                        xmlNode *var_node;
-                        
-                        if (strcmp ((char *) node->name, "property") != 0)
-                                continue;
-                        
-                        /* property */
-                        for (var_node = node->children; var_node;
-                             var_node = var_node->next) {
-                                NotifyData *data;
-                                GValue value = {0, };
-                                GList *l;
-                                
-                                data = g_hash_table_lookup
-                                        (proxy->priv->notify_hash,
-                                         var_node->name);
-                                if (data == NULL)
-                                        continue;
-                                
-                                /* Make a GValue of the desired type */
-                                g_value_init (&value, data->type);
-                                
-                                if (!gvalue_util_set_value_from_xml_node
-                                                        (&value, var_node)) {
-                                        g_value_unset (&value);
-                                        
-                                        continue;
-                                }
-                                
-                                /* Call callbacks */
-                                for (l = data->callbacks; l; l = l->next) {
-                                        CallbackData *callback_data;
-                                        
-                                        callback_data = l->data;
-                                        
-                                        callback_data->callback
-                                                (proxy,
-                                                 (const char *) var_node->name,
-                                                 &value,
-                                                 callback_data->user_data);
-                                }
-                                
-                                /* Cleanup */
-                                g_value_unset (&value);
-                        }
+        if (proxy->priv->sid == NULL)
+                /* No SID */
+                if (G_LIKELY (proxy->priv->subscribed))
+                        /* subscription in progress, delay emision! */
+                        return TRUE;
+
+        for (pending_notify = proxy->priv->pending_notifies;
+             pending_notify != NULL;
+             pending_notify = pending_notify->next) {
+                EmitNotifyData *emit_notify_data;
+
+                emit_notify_data = pending_notify->data;
+
+                if (emit_notify_data->seq > proxy->priv->seq) {
+                        /* Oops, we missed a notify. Resubscribe .. */
+                        resubscribe = TRUE;
+
+                        break;
                 }
-                
-                /* Cleanup */
-                xmlFreeDoc (doc);
+
+                /* Increment our own event sequence number */
+                if (proxy->priv->seq < G_MAXINT32)
+                        proxy->priv->seq++;
+                else
+                        proxy->priv->seq = 1;
+
+                if (G_LIKELY (proxy->priv->sid != NULL &&
+                              strcmp (emit_notify_data->sid,
+                                      proxy->priv->sid) == 0))
+                        /* Our SID, entertain! */
+                        emit_notifications_for_doc (proxy,
+                                                    emit_notify_data->doc);
+        }
+
+        /* Cleanup */
+        while (proxy->priv->pending_notifies != NULL) {
+                emit_notify_data_free (proxy->priv->pending_notifies->data);
 
                 proxy->priv->pending_notifies =
                         g_list_delete_link (proxy->priv->pending_notifies,
                                             proxy->priv->pending_notifies);
         }
-        
+
         proxy->priv->notify_idle_src = NULL;
 
-	return FALSE;
+        if (resubscribe) {
+                unsubscribe (proxy);
+                subscribe (proxy);
+        }
+
+        return FALSE;
 }
 
 /*
@@ -1418,6 +1730,7 @@ server_handler (SoupServer        *soup_server,
         int seq;
         xmlDoc *doc;
         xmlNode *node;
+        EmitNotifyData *emit_notify_data;
 
         proxy = GUPNP_SERVICE_PROXY (user_data);
 
@@ -1428,7 +1741,7 @@ server_handler (SoupServer        *soup_server,
                 return;
         }
 
-        hdr = soup_message_headers_get (msg->request_headers, "NT");
+        hdr = soup_message_headers_get_one (msg->request_headers, "NT");
         if (hdr == NULL || strcmp (hdr, "upnp:event") != 0) {
                 /* Proper NT header lacking */
                 soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
@@ -1436,7 +1749,7 @@ server_handler (SoupServer        *soup_server,
                 return;
         }
 
-        hdr = soup_message_headers_get (msg->request_headers, "NTS");
+        hdr = soup_message_headers_get_one (msg->request_headers, "NTS");
         if (hdr == NULL || strcmp (hdr, "upnp:propchange") != 0) {
                 /* Proper NTS header lacking */
                 soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
@@ -1444,17 +1757,7 @@ server_handler (SoupServer        *soup_server,
                 return;
         }
 
-        hdr = soup_message_headers_get (msg->request_headers, "SID");
-        if (hdr == NULL ||
-            (proxy->priv->sid == NULL ||
-             strcmp (hdr, proxy->priv->sid) != 0)) {
-                /* No SID or not ours */
-                soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
-
-                return;
-        }
-
-        hdr = soup_message_headers_get (msg->request_headers, "SEQ");
+        hdr = soup_message_headers_get_one (msg->request_headers, "SEQ");
         if (hdr == NULL) {
                 /* No SEQ header */
                 soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
@@ -1463,23 +1766,15 @@ server_handler (SoupServer        *soup_server,
         }
 
         seq = atoi (hdr);
-        if (seq > proxy->priv->seq) {
-                /* Oops, we missed a notify. Resubscribe .. */
-                unsubscribe (proxy);
-                subscribe (proxy);
 
-                /* Message was OK otherwise */
-                soup_message_set_status (msg, SOUP_STATUS_OK);
+        hdr = soup_message_headers_get_one (msg->request_headers, "SID");
+        if (hdr == NULL) {
+                /* No SID */
+                soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
 
                 return;
         }
 
-        /* Increment our own event sequence number */
-        if (proxy->priv->seq < G_MAXINT32)
-                proxy->priv->seq++;
-        else
-                proxy->priv->seq = 1;
-
         /* Parse the actual XML message content */
         doc = xmlRecoverMemory (msg->request_body->data,
                                 msg->request_body->length);
@@ -1503,31 +1798,33 @@ server_handler (SoupServer        *soup_server,
 
                 return;
         }
-	
-	/*
-	 * Some UPnP stacks (hello, myigd/1.0) block when sending a NOTIFY, so
-	 * call the callbacks in an idle handler so that if the client calls the
-	 * device in the notify callback the server can actually respond.
-	 */
-        proxy->priv->pending_notifies =
-                g_list_append (proxy->priv->pending_notifies, doc);
-        if (!proxy->priv->notify_idle_src) {
-	        GUPnPContext *context;
-	        GMainContext *main_context;
 
-	        context = gupnp_service_info_get_context
+        /*
+         * Some UPnP stacks (hello, myigd/1.0) block when sending a NOTIFY, so
+         * call the callbacks in an idle handler so that if the client calls the
+         * device in the notify callback the server can actually respond.
+         */
+        emit_notify_data = emit_notify_data_new (hdr, seq, doc);
+
+        proxy->priv->pending_notifies =
+                g_list_append (proxy->priv->pending_notifies, emit_notify_data);
+        if (!proxy->priv->notify_idle_src) {
+                GUPnPContext *context;
+                GMainContext *main_context;
+
+                context = gupnp_service_info_get_context
                         (GUPNP_SERVICE_INFO (proxy));
-	        main_context = gssdp_client_get_main_context
+                main_context = gssdp_client_get_main_context
                         (GSSDP_CLIENT (context));
 
-	        proxy->priv->notify_idle_src = g_idle_source_new();
-	        g_source_set_callback (proxy->priv->notify_idle_src,
-				       emit_notifications,
+                proxy->priv->notify_idle_src = g_idle_source_new();
+                g_source_set_callback (proxy->priv->notify_idle_src,
+                                       emit_notifications,
                                        proxy, NULL);
-	        g_source_attach (proxy->priv->notify_idle_src, main_context);
+                g_source_attach (proxy->priv->notify_idle_src, main_context);
 
                 g_source_unref (proxy->priv->notify_idle_src);
-	}
+        }
         
         /* Everything went OK */
         soup_message_set_status (msg, SOUP_STATUS_OK);
@@ -1638,7 +1935,8 @@ subscribe_got_response (SoupSession       *session,
                 int timeout;
 
                 /* Save SID. */
-                hdr = soup_message_headers_get (msg->response_headers, "SID");
+                hdr = soup_message_headers_get_one (msg->response_headers,
+                                                    "SID");
                 if (hdr == NULL) {
                         error = g_error_new
                                         (GUPNP_EVENTING_ERROR,
@@ -1651,8 +1949,8 @@ subscribe_got_response (SoupSession       *session,
                 proxy->priv->sid = g_strdup (hdr);
 
                 /* Figure out when the subscription times out */
-                hdr = soup_message_headers_get (msg->response_headers,
-                                                "Timeout");
+                hdr = soup_message_headers_get_one (msg->response_headers,
+                                                    "Timeout");
                 if (hdr == NULL) {
                         g_warning ("No Timeout in SUBSCRIBE response.");
 
@@ -1685,13 +1983,13 @@ subscribe_got_response (SoupSession       *session,
 
                         /* Add actual timeout */
                         proxy->priv->subscription_timeout_src =
-			        g_timeout_source_new_seconds (timeout);
-			g_source_set_callback
+                                g_timeout_source_new_seconds (timeout);
+                        g_source_set_callback
                                 (proxy->priv->subscription_timeout_src,
                                  subscription_expire,
                                  proxy, NULL);
-			g_source_attach (proxy->priv->subscription_timeout_src,
-					 main_context);
+                        g_source_attach (proxy->priv->subscription_timeout_src,
+                                         main_context);
 
                         g_source_unref (proxy->priv->subscription_timeout_src);
                 }
@@ -1706,6 +2004,13 @@ subscribe_got_response (SoupSession       *session,
                                  msg->reason_phrase);
 
 hdr_err:
+                /* Remove listener */
+                context = gupnp_service_info_get_context
+                                        (GUPNP_SERVICE_INFO (proxy));
+
+                server = gupnp_context_get_server (context);
+                soup_server_remove_handler (server, proxy->priv->path);
+
                 proxy->priv->subscribed = FALSE;
 
                 g_object_notify (G_OBJECT (proxy), "subscribed");
@@ -1717,13 +2022,6 @@ hdr_err:
                                error);
 
                 g_error_free (error);
-
-                /* Remove listener */
-                context = gupnp_service_info_get_context
-                                        (GUPNP_SERVICE_INFO (proxy));
-
-                server = gupnp_context_get_server (context);
-                soup_server_remove_handler (server, proxy->priv->path);
         }
 }
 
@@ -1874,7 +2172,7 @@ unsubscribe (GUPnPServiceProxy *proxy)
 }
 
 /**
- * gupnp_service_proxy_set_subscribed
+ * gupnp_service_proxy_set_subscribed:
  * @proxy: A #GUPnPServiceProxy
  * @subscribed: %TRUE to subscribe to this service
  *
@@ -1905,7 +2203,7 @@ gupnp_service_proxy_set_subscribed (GUPnPServiceProxy *proxy,
 }
 
 /**
- * gupnp_service_proxy_get_subscribed
+ * gupnp_service_proxy_get_subscribed:
  * @proxy: A #GUPnPServiceProxy
  *
  * Returns if we are subscribed to this service.
